#include "dataconversion.h"

gsl_matrix* cv2gsl_matrix(CvMat* mat)
{
	gsl_matrix* m = gsl_matrix_alloc( mat->rows, mat->cols );

	for(int i=0; i < mat->rows; i++)
		for(int j=0; j < mat->cols; j++)
			// The following works faster, but one must be 
			// careful in doing the right type-cast.
			// Can be dangerous to use if the passed data type
			// is not the right one.
			// m->data[j+i*m->tda] = (double) ((float *) mat->data.ptr)[j+i*mat->cols];	
			// The following is slower but safer, as data-type
			// checking is done by the functions.
			gsl_matrix_set( m, i, j, cvmGet( mat, i, j ) );
			
	return m;
}

CvMat* gsl2cv_matrix(gsl_matrix* m)
{
	CvMat* mat = cvCreateMat( m->size1, m->size2, CV_32FC1 );

	for(int i=0; i < m->size1; i++)
		for(int j=0; j < m->size2; j++)
			// The following works faster, but one must be 
			// careful in doing the right type-cast.
			// Can be dangerous to use if the passed data type
			// is not the right one.
			// ((float *) mat->data.ptr)[j+i*mat->cols] = (float) m->data[j+i*m->tda];	
			// The following is slower but safer, as data-type
			// checking is done by the functions.
			cvmSet( mat, i, j, gsl_matrix_get(m, i, j ) );

	return mat;
}

int cvMatF1toImageF1(CvMat* mat, CvImage* img)
{

	((IplImage *) (*img))->imageData	= (char *) (mat->data.ptr);
	return 0;
}

// This function copies an unsigned char image with 1, 2 or 3 channels to
// an float image.  The function resets the ROI.
int cvImageU3toF3(CvImage* img, CvImage* img_out, bool normalize)
{
	double min_val =  FLT_MAX,
		   max_val = -FLT_MAX;
	int Nc=3; // Number of channels
	unsigned char* ptr_u; // Req. by approach 6
	float* ptr_f;         // Req. by approach 6
	unsigned char alignment_bytes; // Req. by approach 6

	// Compute de alignment bytes (Req. by approach 6)
	// if number not multiple of 4, then number/4 has a mantissa
	// and the floating point division results in a number
	// larger than the integer
	// 
	alignment_bytes = 0;
	if ( (double) (img->width()*img->channels()) /4.0 > 
		 (double) ( img->width()*img->channels()/4 ) )
	{
		alignment_bytes = 4*(img->width()*img->channels()/4)+4-img->width()*img->channels();
	}

	// For CvImage Instance;
	// Nc = img_out.channels(); 
	// For CvImage Pointer
	Nc = img->channels();

	img->reset_roi();
	img_out->reset_roi();

	if (normalize)
	{
	// --- Copy and normalize the image ---
	// MinMaxLoc Does NOT work on float (32 bit) 3-channel images,
	// but works on single channel images.  See normalization of mask below.
	// Does not work: cvSetImageCOI( (IplImage *) &img_out, 1 );
	// Does not work: cvMinMaxLoc( (IplImage *) &img_out, &min_val, &max_val, NULL, NULL, NULL );
	//
	// Note also that cvNormalize does not work because it relies 
	// on cvMinMaxLoc...
	// Does not work: cvNormalize( img_out, img_out, 0.0, 1.0, CV_MINMAX, NULL);
	//
	// So we look for the minimum and maximum values here as the
	// image is copied.
	//
	// For CvImage Instance
	//for(int c = 0; c < Nc; c++)
	//for(int i = 0; i < img.height(); i++)
	//	for(int j = 0; j < img.width(); j++)
	//
	// For CvImage Pointer
	ptr_u = (unsigned char *) (img->data()); // Req. by approach 6
	ptr_f = (float *) (img_out->data()); // Req. by approach 6
	for(int i = 0; i < img->height(); i++)
	{
	for(int j = 0; j < img->width(); j++)
	{	
		for(int c = 0; c < Nc; c++)
		{
			// --- Copy each pixel of img to img_out ---
			// --- Approach 1 ---
			// This is the safest, simplest and slowest method,
			// because it does type and dimension checkings.
			// Avoid it, if you require faster code!!!
			//
			// For CvImage Instance
			// cvSet2D(img_out, i, j, cvGet2D(img, i, j));			
			//
			// For CvImage Pointer
			// cvSet2D((*img_out), i, j, cvGet2D((*img), i, j));
			//
			// --- Approach 2 ---
			// This is the approach proposed in the CV FAQ.
			// This approach is equivalent to approach 3,
			// but employs the IplImage structure instead
			// of the CvImage structure, so it requires
			// additional type-casting.
			//
			// For CvImage Instance
			// ((float*)( ((IplImage *) img_out)->imageData + 
			// 	       ((IplImage *) img_out)->widthStep*i
			// 	      )
			// 	  )[j*Nc + c]=
			// (float) ((unsigned char*)( ((IplImage *) img)->imageData + 
			// 	               ((IplImage *) img)->widthStep*i
			// 	             )
			// 	  )[j*Nc + c];
			//
			// For CvImage Pointer
			// ((float*)( ((IplImage *) (*img_out))->imageData + 
			// 	       ((IplImage *) (*img_out))->widthStep*i
			//	      )
			// 	  )[j*Nc + c]=
			// (float) ((unsigned char*)( ((IplImage *) (*img))->imageData + 
			// 	               ((IplImage *) (*img))->widthStep*i
			// 	             )
			//	  )[j*Nc + c];
			//
			// --- Approach 3 ---
			// This is the fastest approach.
			// Approach 5 is equally fast because it relies
			// on a macro that implements this approach.
			// CvImage's "pix_size()" is the number of bytes per pixel
			//   pix_size() = Number_of_Channels*sizeof(element_type)
			// CvImage's "depth()" is the number of bits per pixel
			//   depth = Number_of_Channels*sizeof(element_type)*8
			//   where sizeof(element_type) returns the number of
			//   bytes of element_type
			// CvImage's "step()" is equivalente to IplImage's 
			//   ".widthStep", which is the number of bytes per pixel
			//   times the number of channels times the number of
			//   pixels per row, i.e.
			//   step() = img.pix_size()*img.width();
			//
			// For CvImage Instance 
			// ((float*)( img_out.data() + img_out.step()*i )
			//	  )[j*Nc + c]=
			// (float) ((unsigned char*)( img.data() + img.step()*i )
			//	  )[j*Nc + c];
			//
			// For CvImage Pointer 
			// ((float*)( img_out->data() + img_out->step()*i )
			//	  )[j*Nc + c]=
			// (float) ((unsigned char*)( img->data() + img->step()*i )
			//	  )[j*Nc + c];
			//
			// --- Approach 4 ---
			// This approach is similar to approach 3, but slower because
			// it involves 3 multiplications and 2 additions instead of
			// 2 multiplications and 2 additions.
			// This approach is easier to read, though.
			//
			// For CvImage Instance
			// ((float*) img_out.data())[j*Nc+c+Nc*img_out.width()*i]=
			// (float) ((unsigned char*) img.data())[j*Nc+c+Nc*img.width()*i];
			//
			// For CvImage Pointer
			// ((float*) img_out->data())[j*Nc+c+Nc*img_out->width()*i]=
			// (float) ((unsigned char*) img->data())[j*Nc+c+Nc*img->width()*i];
			//
			// --- Approach 5 ---			
			// This macro is equivalent to approach 3 (the fastest approach)
			//
			// For CvImage Instance
			// CV_IMAGE_ELEM( (IplImage *) img_out, float, i, j*3+c ) =
			// (float) CV_IMAGE_ELEM( (IplImage *) img, unsigned char, i, j*3+c );
			//
			// For CvImage Pointer
			// CV_IMAGE_ELEM( (IplImage *) (*img_out), float, i, j*Nc+c ) =
			// (float) CV_IMAGE_ELEM( (IplImage *) (*img), unsigned char, i, j*Nc+c );

		    // --- Approach 6 ---
			// Using pointer arithmetic
				*ptr_f = (float) *ptr_u++;
			
			// --- Find min and max values ---
			// --- Approaches 1-5 ---
			// For CvImage Instance
			//if ( CV_IMAGE_ELEM( (IplImage *) img_out, float, i, j*Nc+c ) > max_val )
			//	max_val = CV_IMAGE_ELEM( (IplImage *) img_out, float, i, j*Nc+c );
			//if ( CV_IMAGE_ELEM( (IplImage *) img_out, float, i, j*Nc+c ) < min_val )
			//	min_val = CV_IMAGE_ELEM( (IplImage *) img_out, float, i, j*Nc+c );
			//
			// For CvImage Pointer
			//if ( CV_IMAGE_ELEM( (IplImage *) (*img_out), float, i, j*Nc+c ) > max_val )
			//	max_val = CV_IMAGE_ELEM( (IplImage *) (*img_out), float, i, j*Nc+c );
			//if ( CV_IMAGE_ELEM( (IplImage *) (*img_out), float, i, j*Nc+c ) < min_val )
			//	min_val = CV_IMAGE_ELEM( (IplImage *) (*img_out), float, i, j*Nc+c );
			// --- Approach 6 ---
			if ( *ptr_f > max_val )
				max_val = *ptr_f;
			if ( *ptr_f < min_val )
				min_val = *ptr_f;

			ptr_f++;
		} // c
	} // j
	// --- Approach 6 ---
	// Shift pointer according to the required amount of alignment bytes.
	// This amount is required because the step (Nchannels*ImageWidth) is
	// always approximated to the nearest multiple of 4.
	// Note that the pointer ptr_f to a float does not require any shifting
	// because it is already a multiple of 4 (sizeof(float)=4).
	
		ptr_u+=alignment_bytes;
	} // i

	//printf("min: %f, max: % f\n", min_val, max_val);
	//printf("pixsize: %i\n", (*img_out).pix_size() );
	//printf("step: %i\n", (*img_out).step() );

	// For CvImage Instance
	//cvConvertScale( img_out, img_out, 1.0/(max_val-min_val), 0 ); // ***
	//cvAddS( img_out, cvScalarAll(-min_val/(max_val-min_val)), img_out, NULL ); // ***
	//
	// For CvImage Pointer
	cvConvertScale( (*img_out), (*img_out), 1.0/(max_val-min_val), 0 ); // ***
	cvAddS( (*img_out), cvScalarAll(-min_val/(max_val-min_val)), (*img_out), NULL ); // ***

	} // end if normalize
	else
	{ // copy without normalization
	ptr_u = (unsigned char *) (img->data()); // Req. by approach 6
	ptr_f = (float *) (img_out->data()); // Req. by approach 6
	for(int i = 0; i < img->height(); i++)
	{
	for(int j = 0; j < img->width(); j++)
	{
		for(int c = 0; c < Nc; c++)
		{
			 //CV_IMAGE_ELEM( (IplImage *) (*img_out), float, i, j*Nc+c ) =
			 //(float) CV_IMAGE_ELEM( (IplImage *) (*img), unsigned char, i, j*Nc+c );
			*ptr_f++ = (float) *ptr_u++;
		} // c
	} // j
	// --- Approach 6 ---
	// Shift pointer according to the required amount of alignment bytes.
	// This amount is required because the step (Nchannels*ImageWidth) is
	// always approximated to the nearest multiple of 4.
	// Note that the pointer ptr_f to a float does not require any shifting
	// because it is already a multiple of 4 (sizeof(float)=4).
	
		ptr_u+=alignment_bytes;
	} // i
	} // end else normalize

	return 0;
}

int cvImageNormalizeF3(CvImage* img, CvImage* img_out)
{
	double min_val =  FLT_MAX,
		   max_val = -FLT_MAX;
	int Nc=3; // Number of channels
	float* ptr_f;

	Nc = img_out->channels();

	img->reset_roi();

	ptr_f = (float *) (img->data());
	for(int i = 0; i < img->height(); i++)
	for(int j = 0; j < img->width(); j++)
		for(int c = 0; c < Nc; c++)
		{
			// --- Find min and max values ---
			//if ( CV_IMAGE_ELEM( (IplImage *) (*img), float, i, j*Nc+c ) > max_val )
			//	max_val = CV_IMAGE_ELEM( (IplImage *) (*img), float, i, j*Nc+c );
			//if ( CV_IMAGE_ELEM( (IplImage *) (*img), float, i, j*Nc+c ) < min_val )
			//	min_val = CV_IMAGE_ELEM( (IplImage *) (*img), float, i, j*Nc+c );
			if ( *ptr_f > max_val )
				max_val = *ptr_f;
			if ( *ptr_f < min_val )
				min_val = *ptr_f;

			ptr_f++;
		}

	cvConvertScale( (*img), (*img_out), 1.0/(max_val-min_val), 0 ); // ***
	cvAddS( (*img_out), cvScalarAll(-min_val/(max_val-min_val)), (*img_out), NULL ); // ***
	
	return 0;
}