// ---------------------------------------------------------------------
// File: dataconversion.h
// Author: Miguel Torres-Torriti (c) 2009.01.06-2009.02.19 v. 1.2
// Change log:
//	2009.01.06, v.1.0, MTT: First version was created.
//	2009.01.08, v.1.1, MTT: Code was cleaned and comments were added.
//  2009.01.19, v.1.2, MTT: Pointer arithmetic implemented wherever 
//                          possible to speedup computations.
// Description:
//  Data Conversion Utility Functions
//  Provides functions to convert between OpenCV and GSL matrices,
//  or to convert OpenCV images of unsigned char type to float.
//  For details, see the comments below.  
// Remarks:
//  Requires the OpenCV and GSL libraries to be instaled.
//  The libraries and further information is available from
//  the references provided below.
//
//  OpenCV:
//     http://sourceforge.net/projects/opencvlibrary/
//     http://opencv.willowgarage.com/wiki/
//     http://www.intel.com/cd/software/products/asmo-na/eng/302910.htm     
//  GSL   :
//     http://www.gnu.org/software/gsl/
//     http://david.geldreich.free.fr/dev.html
//
// ---------------------------------------------------------------------

#ifndef DATACONVERSION_H
#define DATACONVERSION_H

#include <cv.h>
#include <gsl/gsl_matrix.h>

// Converts an OpenCV matrix to GSL matrix
gsl_matrix* cv2gsl_matrix(CvMat* mat);

// Converts a GSL matrix to OpenCV matrix
CvMat* gsl2cv_matrix(gsl_matrix* m);

// Assigns the data block pointer of a 1-channel 
// matrix of floating point values to the
// data block pointer of a 1-channel image
// of floating point values.
// Matrix dimensions must agree as no checking
// is done.
int cvMatF1toImageF1(CvMat* mat, CvImage* img);
#define CV_MAT_IMG_F1( mat, img )  \
	((img)->imageData = (char *) ((mat)->data.ptr))

// Copies a 3-channel unsigned char (8-bits per channel's pixel)
// to a 3-channel float (4-bytes per channel's pixel).
// By default returns a normalized image with values between 0 and 1,
// unless the third argument is zero.
int cvImageU3toF3(CvImage* img, CvImage* img_out, bool normalize = 1);

// Normalizes a float img.  In-place operation is supported.
int cvImageNormalizeF3(CvImage* img, CvImage* img_out);

#endif

/* --- Informacin General ---
En resumen, las funciones permiten realizar entre otras cosas
operaciones como:

- Convertir de CvMat de 1 canal double o float a 
  una matriz gsl_matrix (que contiene doubles), y viceversa.
- Convertir de CvImage de 3 canales unsigned char (1 byte por elemento)
  a CvImage de 3 canales float (4 bytes por elemento) 
  con y sin escalamiento, y viceversa.

Vea el cdigo dataconversion.cpp para entender como se accede en forma
rpida (directa) a la ubicacin fsica de la memoria del PC donde 
se almacenan los bytes.  Esto se realiza a travs del uso de punteros!
El acceso a travs de punteros aumenta la eficiencia de rutinas que
son intensivas en "for-loops".  Ud. debera evitar usar funciones
como:

CvScalar pixel = cvGet2D( img, i, j); // Extrae valor de pixel
cvSet2D( img, i, j, pixel );          // Asigna valor a pixel

Esto mismo se puede lograr con:

float pixel_i_j_c = (float) ((unsigned char*)( img.data() + img.step()*i ))[j*Nc + c];
((unsigned char*)( img.data() + img.step()*i ))[j*Nc + c] = (unsigned char) pixel_i_j_c;

donde i, j es el pixel en la fila i, columna j, y 
c es el canal (0 Blue, 1 Green, 2 Red) de la imagen o pixel.

NOTA 1: El uso de los punteros requiere de hacer el type-cast correcto.  Si la imagen contiene pixeles cuyos valores son nmeros float, entonces debe emplearse (float *), si en cambio la imagen tiene valores unsigned char, deber realizarse el type-cast con (unsigned char*).  Adicionalmente si se transfiere una variable char a una float, se requerira un type-cast (float).
Observe adems que el direccionamiento a la posicin de memoria
es del tipo:
 pos = canal + columna*numero_canales + fila*numero_canales*ancho_imagen
y no
 pos = columna + fila*ancho_imagen + canal*ancho_imagen*alto_imagen
porque en OpenCV los canales estn entralazados (interlaced) y no
separados por planos B, G, y R.

Para simplificar la expresin de los punteros anteriores, existen
algunas macros definidas en el archivo cxtypes.h de OpenCV, 
que permiten escribir las dos lneas de ejemplo como:

float pixel_i_j_c = (float) CV_IMAGE_ELEM( (IplImage *) img, unsigned char, i, j*3+c );
CV_IMAGE_ELEM( (IplImage *) img, unsigned char, i, j*3+c ) = (unsigned char)  pixel_i_j_c;

NOTA 2: La macro CV_IMAGE_ELEM tal como est definida en cxtypes.h
est hecha para operar sobre variables tipo IplImage.  Por esta razn,
si la imagen se cre usando la clase CvImage (que es ms prctica)
ser necesario hacer un type-cast previo para extraer los elementos
correspondientes de la forma IplImage.

NOTA 3: Uno puede construir la macro que quiera para evitar escribir
expresiones complicadas con punteros.  Por ejemplo, ver la macro 
CV_MAT_IMG_F1 en este archivo, que convierte una matrix de 1 canal
float a una imagen de un canal tambin float.

*** Aritmetica de Punteros y mejoras en la velocidad de procesamiento ***
El metodo mas rapido para acceder elementos de una matriz o imagen que
se barre en forma secuencial es mediante aritmetica de punteros.

Suponga que se desea accesar al elemento i, j de la matriz M de m x n elementos.
Como se explico en los parrafos anteriores, una manera de realizarlo seria
mediante:
	val = M[j+n*i]

Si es necesario acceder a todos los elementos, entonces:
	for(int i=0;i<m;i++)
		for(int j=0;j<n;j++)
			val = M[j+n*i];
El metodo anterior requiere m x n multiplicaciones de los valores n e i, y 
m x n sumas de j mas el termino recien multiplicado n*i.

Las multiplicaciones son operaciones costosas a nivel de la CPU, por lo que
es preferible realizar:
	float* ptr;
	ptr = M;
	for(int i=0;i<m;i++)
		for(int j=0;j<n;j++)
			val = *ptr++;
La linea val = *ptr++ significa, tomar el valor apuntado por ptr, es decir *ptr
y pasarlo a val, luego incrementar el puntero ejecutando ptr++.  En este caso
solo se requiere realizar m x n sumas simples, ya que ptr++ equivale a ptr = ptr+1,
que a nivel de la CPU es una operacion basica de incrementar en 1 bit el valor de ptr.

Al trabajar con imagenes de varios canales es posible lograr reducciones
en el tiempo de computo de 3 a 4 veces!  Esto es altamente significativo
al procesar video ya que podra aumentar la cantidad de cuadros procesados
por segundo!

*/