//MATH FUNCTIONS LIBRARY - M.A.TT. 21.VIII.2000
#include <stdio.h>
#include <stdlib.h>
#include "mathlib.h"
#include "util.h"

// *****************************************************************************
// --- BASIC MATH FUNCTIONS ---
// Miguel Torres Torriti (c) 2008.11.21 v. 1.0
// These functions have been created to illustrate the construction 
// of vectors and matrices.  These functions are improved versions
// of the old Math Library, however it is recommended that the
// versions implemented as part of OpenCV or the GSL libraries
// are used instead for portability.
//

// Functions to create basic mathematical structures (vectors and matrices).

// --- Vector functions ---
//struct mvector{        // mvector structure
//	//size_t dt;       // data type (bytes per element)
//	unsigned long len; // vector size 
//	double *val;	   // vector data values
//};

// Allocate memory for vector
mvector* mvectoralloc(unsigned long len)
{
	mvector *v=(mvector *) malloc(sizeof(mvector));   // allocate mem for the vector structure
	(*v).len=len;
	(*v).val = (double *) malloc(sizeof(double)*len); // allocate mem for the vector's data 
	if ((*v).val == (double *) NULL)
	{
		fprintf(stderr,"mvectoralloc(): Memory allocation error.\n");
		return (mvector *) NULL;
	}
	else
		return (mvector *) v;
}

// Free memory used by the vector
void mvectorfree(mvector *v)
{
	if (v != (mvector *) NULL)
	{
		if ((*v).val != (double *) NULL)
			free((*v).val);
		free(v);
	}
}

// Print the vector's content
void mvectorprint(mvector *v)
{	
	for(unsigned int i=0;i<(*v).len;i++)
		printf("v[%d]: %.2f\n", i, (*v).val[i]);
}

// Fill vector with test values (for testing purposes)
void mvectorfilltest(mvector *v)
{
	for(unsigned int i=0;i<(*v).len;i++)
	{
		(*v).val[i]=(double) i*2.5;
	}
}


// --- Matrix functions ---
//struct mmatrix{         // mmatrix structure
//	unsigned long rows; // number of rows
//	unsigned long cols; // number of columns
//	double **val;	    // matrix data values
//};

// Allocate memory for matrix
mmatrix* mmatrixalloc(unsigned long rows, unsigned long cols)
{
	mmatrix *m=(mmatrix *) malloc(sizeof(mmatrix));     // allocate mem for the vector structure
	(*m).rows=rows;
	(*m).cols=cols;
	
	(*m).val = (double **) malloc(sizeof(double *)*rows); // allocate mem for "a column" of length "rows" 
	                                                    // of pointers to row pointers 
	if ((*m).val == (double **) NULL)
		fprintf(stderr,"mmatrixalloc(): Stage 1 memory allocation error.\n");
	
	double *linarray = (double *) malloc(sizeof(double)*rows*cols); // allocate mem for the matrix as a linear array
	if (linarray == (double *) NULL)
	{
		fprintf(stderr,"mmatrixalloc(): Stage 2 memory allocation error.\n");
		return (mmatrix *) NULL;
	}
	else
	{
		for(unsigned long i=0, j=0; i<rows;i++, j+=cols)
		{	
			(*m).val[i] = linarray+j;
		}
		return (mmatrix *) m;
	}
}

// Free memory used by the matrix
void mmatrixfree(mmatrix *m)
{
	if (m != (mmatrix *) NULL)
	{
		if ((*m).val[0] != (double *) NULL)
			free((*m).val[0]);
		if ((*m).val != (double **) NULL)
			free((double *) ((*m).val));
		free(m);
	}
}

// Print the matrix's content
void mmatrixprint(mmatrix *m)
{	
	for(unsigned int i=0;i<(*m).rows;i++)
	{
		for(unsigned int j=0;j<(*m).cols;j++)
			printf("m[%d,%d]: %.2f ", i, j, (*m).val[i][j]);
		printf("\n");
	}
}

// Fill matrix with test values (for testing purposes)
void mmatrixfilltest(mmatrix *m)
{
	for(unsigned int i=0;i<(*m).rows;i++)
		for(unsigned int j=0;j<(*m).cols;j++)
		{
			(*m).val[i][j]=(double) (i*(*m).cols+j);
		}
}

// *****************************************************************************
// MATH FUNCTIONS OF THE OLD MATH LIBRARY - M.A.TT. 21.VIII.2000
// These functions have been abandoned in favor of similar
// versions implemented as part of OpenCV or the GSL libraries.
//
void *vector(unsigned long nElems, int size)
{
void	*ptr;

	ptr = (void *) malloc (nElems * size);
	if (ptr == (void *) NULL)
		fatalError ("Vector(): Memory allocation error.\n");
	return (ptr);
} /* vector() */


/*
 * Free a 2D array as allocated by matrix().
 *
 * NOTE:  The array name should be cast to type (void **) to avoid
 * compiler warnings.  This is a little naughty of me and should be fixed...
 * sometime...
 *
 */
void freeMat(void **ptr2d)
{
	if (ptr2d[0] != (void *) NULL)
		free ((void *) ptr2d[0]);
	if (ptr2d != (void **) NULL)
		free ((void *) ptr2d);
} /* freeMat() */


/*
 *  Dynamically allocate a 2D (x*y) array of elements of size x*y*size bytes.
 */
void **matrix (unsigned long x, unsigned long y, int size)
{
void    *ptr;
void    **returnPtr;
register unsigned long    i, j;

        // Allocate vector of x double pointers rows
        returnPtr = (void **) malloc (x * sizeof(void *));
	if (returnPtr == (void **) NULL)
		fatalError ("matrix():  Stage 1 memory allocation error.\n");
	ptr = (void *) malloc (x * y * size);
	if (ptr == (void *) NULL)
		fatalError ("matrix():  Stage 2 memory allocation error.\n");
        // Asign a  row of columns to each row pointer
        for (i=0, j=0; j<x; i+=(y*size), j++)
			//MTT20090208   returnPtr[j] = (void *) ((long)ptr + i);
			returnPtr[j] = (void *) ((uchar*) ptr+i);
	return ((void **) returnPtr);
} // matrix()


void copy_matrix(unsigned int rows, unsigned int cols,
                 double **matrix_in, double **matrix_out)
{
 for(unsigned int i=0;i<rows;i++)
 for(unsigned int j=0;j<cols;j++)
         matrix_out[i][j]=matrix_in[i][j];
}

/*
 *  enorm2(): Calculates the squared Euclidean norm of an n-dimensional
 *            vector x.
 *  In this case checking for over/underflows is irrelevant since
 *  overflows they will be unavoidable, even if the squared norm is
 *  first calculated by squaring the result obtained with enorm().
 *  cf. enorm(), below.
 *
 */

double enorm2(double *vec, unsigned int dimVec)
{
 double enorm2=0.0;

 for(unsigned int i=0;i<dimVec;i++)
 {
  enorm2+=vec[i]*vec[i];
 }

 return enorm2;
} // enorm2()


/*
 *  enorm(): Calculates the Euclidean norm of an n-dimensional vector x.
 *
 *  Function translated from the original version written in Fortran by:
 *   Burton S. Garbow, Kenneth E. Hillstrom, Jorge J. More
 *   Argonne National Laboratory. Minpack Project. March 1980.
 *
 *  Original authors description:
 *    the euclidean norm is computed by accumulating the sum of
 *    squares in three different sums. the sums of squares for the
 *    small and large components are scaled so that no overflows
 *    occur. non-destructive underflows are permitted. underflows
 *    and overflows do not occur in the computation of the unscaled
 *    sum of squares for the intermediate components.
 *    the definitions of small, intermediate and large components
 *    depend on two constants, rdwarf and rgiant. the main
 *    restrictions on these constants are that rdwarf**2 not
 *    underflow and rgiant**2 not overflow. the constants
 *    given here are suitable for every known computer.
 *
 */

double enorm(double *vec, unsigned int dimVec)
{
 double enorm=0.0;
 double rdwarf=3.834e-20, rgiant=1.304e+19;
 double s1=0.0, s2=0.0, s3=0.0, x1max=0.0, x3max=0.0, agiant, xabs;

 agiant = rgiant/(double) dimVec;

 for(unsigned int i=0;i<dimVec;i++)
 {
    xabs = fabs(vec[i]);
    if ((rdwarf < xabs)&&(xabs < agiant))
    {
       // Sum for intermediate components
       s2+=xabs*xabs;
    }
    else if (xabs <= rdwarf)
    {
       // Sum for small components
       if (xabs > x3max)
       {
          s3 = 1.0 + s3*(x3max/xabs)*(x3max/xabs);
          x3max = xabs;
       }
       else
       {
          if (xabs!=0.0)
             s3+=(x3max/xabs)*(x3max/xabs);
       }
    }
    else
    {
       // Sum for large components
       if (xabs > x1max)
       {
          s1 = 1.0 + s1*(x1max/xabs)*(x1max/xabs);
          x1max = xabs;
       }
       else
       {
          s1+= (xabs/x1max)*(xabs/x1max);
       }
    }
 } // for i

 // Calculation of the norm
 if (s1==0.0)
 {
    if (s2==0.0)
    {
        enorm = x3max*sqrt(s3);
    }
    else
    {
       if (s2<x3max)
          enorm = sqrt(x3max*((s2/x3max)+(x3max*s3)));
       else
          enorm = sqrt(s2*(1.0+(x3max/s2)*(x3max*s3)));
    }
 }
 else
 {
    enorm = x1max*sqrt(s1+(s2/x1max)/x1max);
 }

 return enorm;
} // enorm()


/*
 *
 * The function row_mean() and covar_matrix() take as input a matrix of
 * column vectors called **vectors, assuming that each column an observation
 * and each row a variable.  Hence, the covariance matrix will be
 * of size rows x rows.
 *
 */

void row_mean(unsigned int rows, unsigned int cols,
              double **vectors, double *mean)
{
unsigned int	i, j;
double	sum;

	for (i=0; i<rows; i++)
	{
	     sum = 0.0;
	     for (j=0; j<cols; j++)
	     {
		  sum += vectors[i][j];
             }
	mean[i] = sum / (double) cols;
	}
} // row_mean()



void covar_matrix (int rows, int cols, double **vectors,
		   double *mean, double **covariance)
{
int	i, j, k;
double	sum;

	for (i=0; i<rows; i++)
            {for (j=0; j<rows; j++)
	         {sum = 0.0;
		  for (k=0; k<cols; k++)
                      {
                       // EVALUATE:  [Xi-mean(Xi)]^T*[Xj-mean(Xj)]

                       sum += (vectors[i][k] - mean[i]) *
				(vectors[j][k] - mean[j]);
                      }
		   covariance[i][j] = sum / (double) cols;
		  }
	    }
} // covar_matrix()

/* 
 *	Function:	ludcmp()
 *
 *	Given a matrix a[0..(n-1)][0..(n-1)], this routine replaces it by the LU 
 *	decomposition of a rowwise permutation of itself. a and n are input. 
 *	a is output, arranged as in equation (2.3.14) above;
 *	indx[0..(n-1)] is an output vector that records the row permutation 
 *	effected by the partial pivoting; d is output as +/- 1 depending on 
 *	whether the number of row interchanges was even or odd, respectively. 
 *	This routine is used in combination with lubksb to solve linear 
 *	equations or invert a matrix.
 *
 *	Acknowledgment:  William H. Press, "Numerical Recipes in C: the art of
 *	scientific computing", 2nd edition, Cambridge University Press, 
 *	Cambridge, 1992
 */
#define TINY 1.0e-20;	/* A small number. */

void ludcmp(double **a, int n, int *indx, double *d)
{
int		i,imax,j,k;
double	big,dum,sum,temp;
double	*vv; 	/* vv stores the implicit scaling of each row. */

	imax = 0;
	vv = (double *) vector(n, sizeof(double));
	*d=1.0;			/* No row interchanges yet. */
/* 
 *	Loop over rows to get the implicit scaling information. 
 */
	for (i=0; i<n; i++) { 
		big=0.0;
		for (j=0; j<n; j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) fatalError("Singular matrix in routine ludcmp");
	/* No nonzero largest element. */
		vv[i]=1.0/big;	/* Save the scaling. */
	}
	for (j=0; j<n; j++) {	/* the loop over columns of Crout's method. */
		for (i=0; i<j; i++) {	/* equation (2.3.12) except for i=j */
			sum=a[i][j];
			for (k=0; k<i; k++) sum -= a[i][k]*a[k][j];
				a[i][j]=sum;
		}
		big=0.0;	/* Initialize search for largest pivot element. */
		for (i=j;i<n;i++) { /* i=j of equation (2.3.12) and i=j+1...N */
							/* of equation (2.3.13). */
			sum=a[i][j];
			for (k=0; k<j; k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
/* 
 * Is the figure of merit for the pivot better than the best so far?
 */
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) { /*	do we need to interchange rows? */
   			for (k=0; k<n; k++) { /* Yes, do so... */
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);		/* ...and change the parity of d. */
			vv[imax]=vv[j];	/* Also interchange the scale factor. */
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
/*
 *	If the pivot element is zero the matrix is singular (at least to the 
 *	precision of the algorithm). For some applications on singular 
 *	matrices, it is desirable to substitute TINY for zero.
 */
		if (j < (n-1)) {	
/* 
 *	Now, finally, divide by the pivot element. 
 */
			dum=1.0/(a[j][j]);
			for (i=j+1; i<n; i++) a[i][j] *= dum;
		}
	}	/* Go back for the next column in the reduction. */
	free(vv);
} /* ludcmp() */


/*
 *	Function: lubksb()
 *
 *	Solves the set of n linear equations A X = B. Here a[0..(n-1)][0..(n-1)] 
 *	is input, not as the matrix A but rather as its LU decomposition, 
 *	determined by the routine ludcmp. indx[0..(n-1)] is input as the 
 *	permutation vector returned by ludcmp. b[0..(n-1)] is input as the 
 *	right-hand side vector B, and returns with the solution vector X. 
 *	a, n, and indx are not modified by this routine and can be left in place 
 *	for successive calls with different right-hand sides b. This routine 
 *	takes into account the possibility that b will begin with many zero 
 *	elements, so it is efficient for use in matrix inversion.
 *
 *	Acknowledgment:  William H. Press, "Numerical Recipes in C: the art of
 *	scientific computing", 2nd edition, Cambridge University Press, 
 *	Cambridge, 1992
 */
void lubksb(double **a, int n, int *indx, double *b)
{
int		i,ii,ip,j;
double	sum;

/* 
 *	When ii is set to a positive value, it will become the index of 
 *	the first nonvanishing element of b. We now do the forward substitution, 
 *	equation (2.3.6). The only new wrinkle is to unscramble the permutation
 *	as we go.
 */
	ii=-1;
	for (i=0; i<n; i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii >= 0)
			for (j=ii; j<i; j++) sum -= a[i][j]*b[j];
		else if (sum) 
			ii=i; 
/*
 *	A nonzero element was encountered, so from now on we will have to 
 *	do the sums in the loop above.
 */
		b[i]=sum;
	} 

/* 
 *	Now we do the backsubstitution, equation (2.3.7).
 */
	for (i=n-1; i>=0; i--) {
		sum=b[i];
		for (j=i+1; j<n; j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];  /* Store a component of the solution vector X */
	}
} /* lubksb() */

/*
 *
 * Calculates the inverse of a matrix using LU decomposition and
 * back substitution (ludcmp() and lubksb()).
 *
 */  
void matrix_inv (int n, double **matrix_in, double **inv)
{
int	i, j; //indx[n];
int *indx;
double	permusign, **a; //col[n]
double *col;
indx = new int[n];
col = new double[n];
a=(double **) matrix(n,n,sizeof(double));
copy_matrix(n,n,matrix_in,a);

/*
 *	Decompose the matrix in LU, both L and U returned in a
 */
	ludcmp(a, n, indx, &permusign);

/*
 *	Find inverse by columns
 */
	for (j=0; j<n; j++) {
		for (i=0; i<n; i++)
			col[i] = 0.0;
		col[j] = 1.0;
		lubksb(a, n, indx, col);
		for (i=0; i<n; i++) inv[i][j] = col[i];
	}
	//free (indx);
	//free (col);
        freeMat((void **) a);
		delete [] indx;
		delete [] col;
} // matrix_inv()

/*
 *
 * Calculates the determinant of a matrix using LU decomposition and
 * back substitution (ludcmp() and lubksb()).
 *
 */  
void matrix_det (int n, double **matrix_in, double *det)
{
int     k; //indx[n];
int *indx;
double **a;
indx = new int[n];
a=(double **) matrix(n,n,sizeof(double));
copy_matrix(n,n,matrix_in,a);

/*
 *	Decompose matrix and return determinant's sign as +/- 1
 */
	ludcmp(a, n, indx, det);

/*
 *	Find determinant
 */
	for(k=0;k<n;k++)
		*det *= a[k][k];

	//free (indx);
        freeMat((void **) a);
		delete [] indx;

} // matrix_det()

/*
 *
 * Calculates the inverse and the determinant of a matrix using
 * LU decomposition and back substitution (ludcmp() and lubksb()).
 *
 */  
void matrix_inv_det (int n, double **matrix_in, double **inv, double *det)
{
int	i, j; //indx[n];
int *indx;
double	**a; //col[n];
double *col;
indx = new int[n];
col = new double[n];
a=(double **) matrix(n,n,sizeof(double));
copy_matrix(n,n,matrix_in,a);

/*
 *	Decompose the matrix in LU, both L and U returned in a
 */
	ludcmp(a, n, indx, det);

/*
 *	Find inverse by columns
 */
	for (j=0; j<n; j++) {
		for (i=0; i<n; i++)
			col[i] = 0.0;
		col[j] = 1.0;
		lubksb(a, n, indx, col);
		for (i=0; i<n; i++) inv[i][j] = col[i];
	}

/*
 *	Find determinant
 */
	for(i=0;i<n;i++)
		*det *= a[i][i];

        //free (indx);
	//free (col);
        freeMat((void **) a);
		delete [] indx;
		delete [] col;
} // matrix_inv()


/**** END OF MATHLIB.CPP ****/
