/* PTZ - Project Template for MS Visual Studio .Net 2003 (Visual C++ 7)
 * Miguel Torres Torriti (c) 2008.11.21-2009.02.06 v. 2.0
 * ---------------------------------------------------------------------
 *
 * To run type:
 *   ptz 
 *   ptz [camera_number] 
 *   ptz <video_filename>
 *   ptz <base_filename> [options] [output_filename=base_filename.jpg]
 *
 * For examples included:
 *   ptz                   -> captures video from camera if connected, otherwise loads image 'input.jpg'
 *   ptz video.avi         -> captures video from video file 'video.avi'
 *   ptz f -e jpg -sq 0 10 -> loads files f000.jpg, f001.jpg ... f010.jpg
 *   ptz input             -> loads file input.jpg
 *
 * Description: 
 *  Opens a file, a file sequence, a video file or video stream from a camera, shows the images
 *  and applies a simple edge detection process to the images.
 *
 * Command-line Options:
 *  -h  Prints some help.
 *  -nv Turns verbose off.
 *  -np Do not wait for user to press ESC when done.
 *  -nd Do not display images.
 *  -no Do not save the output.
 *  -e  Filename extension [=jpg]
 *  -sq initial_value final_value
 *         Tells ptz to process a sequence of images
 *         starting at filename<initial_value>.ext and
 *         ending in filename<initial_value>.ext
 *         By default the sequence range is assumed to be between 0 and 10
 *         with zero-padding = 3, i.e. 000, 001, ..., 010.
 *  -zp num  Pad filename suffix with num zeros.
 *  -fps num  Frames per seconds [=max_fps].
 *
 * Keyboard inputs:
 *  ESC - quit the program
 *  Mouse-R - select region
 *  Mouse-R+CTRL - draw on buffer
 *  p - process current image using loaded or precomputed data
 *  a - add class
 *  r - remove current class
 *  l - list existing classes
 *  u - calculate feature vector for current region and update the class parameters
 *  b - blank current class data
 *  v - fill current class data with random values
 *  < - decrease current class
 *  > - increase current class
 *  h - toggle process holding (temporarily disables/enables processing)
 *  g - get class data from cpd.txt
 *  s - save class data to cpd.txt
 *  t - test basic functions
 *
 * Remarks:
 *  The main purpose of this code is to provide a template for writing image processing routines
 *  relying on OpenCV's functionality.  The template code is structured in terms of the typical 
 *  steps involved in the implementation of an image processing solution:
 *
 *		1. Initialize data structures
 *		2. Parse command line parameters and update data structures
 *		3. Get the input image
 *			Depending on the context some of the typical actions 
 *			may include commands with names such as:
 *				Allocate/Create/Open/Load/Read/Get/Retrieve/Grab/Capture
 *		4. Process image
 *		5. Store the results
 *			Depending on the context some of the typical actions 
 *			may include commands with names such as:
 *				Save/Write/Set/Put/Close/Release/Free/Clear/Destroy
 *		6. Show/display results
 *		7. Handle keyboard inputs from the user
 *		8. Repeat 3-7 until some ending condition is met, skipping 6 and 7 if no user interaction 
 *         through CLI or GUI is required in order to save time.
 *
 *  In addition to providing a structure with a logical flow that should help you keeping solutions
 *  ordered and easier to work upon, the template is also intended to serve as a guide on how 
 *	to comment and document your code.  WHEN WRITING YOUR OWN CODE, THIS HEADING SHOULD BE UPDATED
 *	WITH YOUR OWN EXPLANATIONS AND NOTES. ALSO COMMENTS TO YOUR OWN FUNCTIONS SHOULD BE ADDED.
 *  DO NOT FORGET TO PRESERVE THE CODE STRUCTURE GIVEN IN STEPS 1-8, AND EXPLAIN HOW TO BUILD AND
 *  COMPILE THE PROJECT, STATING WHICH LIBRARIES MUST BE INSTALLED AND LINKED, AS WELL AS OTHER 
 *  PARAMETER DEFINITIONS REQUIRED.
 *
 *  The structures and routines can be divided into five main groups:
 *   1. Basic utility functions
 *   2. Basic math functions 
 *   3. Process functions
 *   4. GUI functions
 *   5. Main
 *  When writing your code, groups 1, 4, and 5 should be updated with your own functions, while
 *  groups 2 and 3 may be completely removed and replaced by your own structures and functions.
 * 
 * Author: Miguel Torres Torriti (c) 2008.11.22, v. 1.
 *
 * Change log:
 *	2008.11.22, v.1.0, MTT: First (single-file) version was created.
 *  2009.02.06, v.2.0, MTT: Source code was split into different .h and .cpp files.
 *	2009.11.22, v.x.x, nnn: If you create a new version, comment here the 
 *                            modifications to the previous one.
 *
 * How to compile:
 *  Set project properties by selecting the menu option "Project/PTZ Properties" as follows:
 * 
 *    a. For Debug version:
 *       C/C++/Additional Include Directories = C:/Program Files/OpenCV/cxcore/include,
 *                                              C:/Program Files/OpenCV/cv/include,
 *                                              C:/Program Files/OpenCV/otherlibs/highgui
 *       C/C++/Code Generation/Multi-threaded Debug DLL (/MDd)
 *       Linker/General/Enable Incremental Linking = YES
 *       Linker/General/Additional Library Directories = C:/Program Files/OpenCV/lib
 *       Linker/Input/Additional Dependecies = cxcored.lib cvd.lib highguid.lib  // for OpenCV
 *                                             gsl_d.lib cblas_d.lib             // for GSL
 *       Linker/Input/Ignore Specific Library = libcd.lib    // for GSL
 *                                              libmmdd.lib  // for OpenCV (Intel Compiler Multithreaded Mathematical Dynamic Library)
 *    b. For Release version:
 *       C/C++/Additional Include Directories = C:/Program Files/OpenCV/cxcore/include,
 *                                              C:/Program Files/OpenCV/cv/include,
 *                                              C:/Program Files/OpenCV/otherlibs/highgui
 *       C/C++/Code Generation/Multi-threaded DLL (/MD)
 *       Linker/General/Enable Incremental Linking = NO
 *       Linker/General/Additional Library Directories = C:/Program Files/OpenCV/lib
 *       Linker/Input/Additional Dependecies = cxcore.lib cv.lib highgui.lib    // for OpenCV 
 *                                             gsl.lib cblas.lib                // for GSL
 *       The following may not be necessary for newer versions of OpenCV / GSL
 *       Linker/Input/Ignore Specific Library = libc.lib    // for GSL
 *                                              libmmd.lib  // for OpenCV (Intel Compiler Multithreaded Mathematical Dynamic Library)
 *  
 * References
 *  OpenCV:
 *     http://sourceforge.net/projects/opencvlibrary/
 *     http://opencv.willowgarage.com/wiki/
 *     http://www.intel.com/cd/software/products/asmo-na/eng/302910.htm     
 *  GSL   :
 *     http://www.gnu.org/software/gsl/
 *     http://david.geldreich.free.fr/dev.html
 *  C/C++ :
 *     http://publications.gbdirect.co.uk/c_book/
 *     http://www.cppreference.com/
 *     http://www.cplusplus.com/
 *     http://www.acm.uiuc.edu/webmonkeys/book/c_guide/
 * 
 *    
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <conio.h>
#include <time.h>
#include "cv.h"       // use " " instead of < > to get list of class/struct members
#include "highgui.h"  // use " " instead of < > to get list of class/struct members
#include "util.h"
#include "mathlib.h"
#include "test.h"
#include "process.h"
#include "gui.h"

// --- MAIN --- ///

int main( int argc, char** argv )
{
	CvCapture* capture = 0;	// Pointer to the OpenCV capture object.
	IplImage* frame = 0;	// Pointer to the capture object frame.
							// Warning: this pointer points to some internal buffer
							// so do not rely on this image as it may be modified 
							// on the next call to grab a frame.  This image must be
							// copied or cloned to elsewhere to work with it.
							// Because this pointer points to a structure maintained
							// by OpenCV itself, you are not required to release it. 
	long int video_length = 0;	// Lenght (total number of frames) of input video file.
	static CvImage img;		// Stores the input image.
	bool hold_flag = 1;		// Hold program from processing the input unless hold_flag=0.
	long int file_num = 0;  // Indicates the current file number for a sequence of files    
	char filename[512];     // Frame filename
	char color_index=0;

	// Test ground
	test_ground( argc, argv );

	// Initialize command line parameters to default values  
	initcmdlinestruct( &cmdlineparam );		

	// Parse the command line parameters 
	if (parsecmdline(argc, argv, &cmdlineparam)==-2)
		return 0;

	// Print welcome message and the OpenCV version
	if (cmdlineparam.verbose_flag)
		printwelcome();

	// Check if input is a live video stream or a video file
    if( argc == 1 || (argc == 2 && strlen(argv[1]) == 1 && isdigit(argv[1][0])))
        capture = cvCaptureFromCAM( argc == 2 ? argv[1][0] - '0' : 0 );
    else if( argc >= 2 )
        capture = cvCaptureFromAVI( argv[1] );

	if( !capture )  // load image from file
    {
        fprintf(stderr,"Could not initialize capturing...\n");
		fprintf(stderr,"Will process file...\n");
        //return -1;
		
		makefilename(&cmdlineparam, filename, cmdlineparam.iv);
		//img.load(argc > 1 ? argv[1] : filename, 0, CV_LOAD_IMAGE_COLOR);
		img.load( filename, 0, CV_LOAD_IMAGE_COLOR);

		if( !img.data() ) // Check if the image has been loaded properly
		{
			fprintf( stderr, "ERROR: NULL image loaded!\n" );
			getchar();
			return -1;
		}
    }
	else // capture from video stream or file
	{
		frame = cvQueryFrame( capture );
		img = cvCloneImage( frame );
		cvSetCaptureProperty( capture, CV_CAP_PROP_POS_FRAMES, 0.0 );
		video_length = (long) cvGetCaptureProperty( capture, CV_CAP_PROP_FRAME_COUNT );
		if (video_length > 0) printf("Video length: %ld frames\n", video_length );
	}

	// Create a window and display the image
	if (cmdlineparam.display_flag)
	{
		cvNamedWindow( wndname, CV_WINDOW_AUTOSIZE );
		img.show( wndname ); // .show method is the conveninient form of cvShowImage	
		cvSetMouseCallback( wndname, on_mouse, (void *) &img );

		cvNamedWindow( wndname2, CV_WINDOW_AUTOSIZE );
		// Create a toolbar 
		cvCreateTrackbar(tbarname2, wndname2, &cr_thresh, 100, on_trackbar2);
		// Show the image
		on_trackbar2(0);

		cvNamedWindow( wndname3, CV_WINDOW_AUTOSIZE );
		// Create a toolbar 
		cvCreateTrackbar(tbarname3, wndname3, &edge_thresh, 2, on_trackbar3);
		// Show the image
		on_trackbar3(0);
	}

	// Initialize/create process output images
	imgIRB = img.clone();							// Create image to store the YCrCb version
	imgSeg = img.clone();							// Create image to store the segmentation of the input image
	imgSegMask.create(img.size(), IPL_DEPTH_8U, 3);	// Create image to store the segmentation mask
	imgG.create(img.size(), IPL_DEPTH_8U, 1);		// Create image to store the Grayscale version

	if (capture)
	{
		((IplImage *) imgSegMask)->origin = img.origin();  // set the origin position to match that of the input
		((IplImage *) imgG)->origin = img.origin();    // set the origin position to match that of the input
	}

	// Initialize process parameters/variables
	cpar.Nc = 0;
	cpar.c = 0;
	cpar.fdim = 3;
	cpar.fv=NULL;

	// --- Process loop begins here ---
	file_num = cmdlineparam.iv;
	//clock_t t0=clock()-(clock_t) ((1.0/cmdlineparam.fps)* (double) CLOCKS_PER_SEC);
	double t0 = (double) clock()/CLOCKS_PER_SEC - 1.0/cmdlineparam.fps;

	if (cmdlineparam.verbose_flag)
	{
		for(;;)
		{
			//if ((double) (clock()-t0)/ (double) CLOCKS_PER_SEC >= 1.0/cmdlineparam.fps)
			if ((double) clock()/CLOCKS_PER_SEC - t0 >= 1.0/cmdlineparam.fps)
			{
				t0 = (double) clock()/CLOCKS_PER_SEC;
		
				// Capture/load images 
				if (capture) // capture from video stream or file
				{
					frame = 0;
					frame = cvQueryFrame( capture );
					//cvCopy( frame, img, NULL); 
					img = cvCloneImage( frame );
					if ((video_length>0) && (cvGetCaptureProperty( capture, CV_CAP_PROP_POS_FRAMES) == video_length))
						break;
				}
				else if (file_num > -1) // load image from file
				{
					makefilename(&cmdlineparam, filename, file_num);
					printf("Loading %s...\n", filename);
					img.load( filename, 0, CV_LOAD_IMAGE_COLOR);
					file_num++;
					if (file_num > cmdlineparam.fv)
						break;
				}

				// Process Input
				if (!hold_flag)
				{
					// Display input image
					img.show( wndname );

					// Process input
					ProcessColorSegmentation(&img, cmdlineparam);
						
					// Display output
					imgSeg.show( wndname2 );
					imgG.show( wndname3 );
				}
				else
				{
					// Display input image
					img.show( wndname );

					// Display unprocessed input in YCrCb
					ProcessColorConversion(&img);
					imgIRB.show( wndname2 );
					imgG.show( wndname3 );
				}
			
			} // end if clock...


			// Process keyboard
			int c_int = cvWaitKey(10);
			if ( c_int != -1 )
			{
				unsigned char c = (unsigned char) c_int;
				if (c == 'p'||c == 'P')
				{
					printf("Processing all classes...\n");
					DetectClasses( &img, &cpar );
					imgSeg.show( wndname2 );
				}
				else if (c == 'a'||c == 'A')
				{
					printf("Adding a new class...\n");
					AddClass( &cpar );
					printf("Num Classes %d\n", cpar.Nc);
				}
				else if (c == 'r'||c == 'R')
				{	
					printf("Removing current class...\n");
					RemoveClass( &cpar );
				}
				else if (c == 'l'||c == 'L')
				{
					printf("Listing classes data...\n");
					PrintClass( &cpar );
				}
				else if (c == 'u'||c == 'U')
				{
					printf("Updating current class parameters...\n");
					UpdateClass( &cpar );
				}
				else if (c == 'v'||c == 'V')
				{
					printf("Randomly filling current class parameters...\n");
					RandomlyFillClass( &cpar );
				}
				else if (c == 'b'||c == 'B')
				{
					printf("Blanking current class parameters...\n");
					BlankClass( &cpar );
				}
				else if (c == '>')
				{
					printf("Incrementing current color index...\n");
					IncrementClass( &cpar );
					color_index++;
					if (color_index>19) color_index=0;
				}
				else if (c == '<')
				{
					printf("Decrementing current color index...\n");
					DecrementClass( &cpar );
					color_index--;
					if (color_index<0) color_index=18;
				}
				else if (c == 'h'||c == 'H')
				{
					printf("Turning off process holding ...\n");
					hold_flag = !hold_flag;
				}
				else if (c == 'g'||c == 'G')
				{
					printf("Getting class parameters data...\n");
					ReadCPD( &cpar, "cpd.txt" );
				}
				else if (c == 's'||c == 'S')
				{
					printf("Saving class parameters data...\n");
					WriteCPD( &cpar, "cpd.txt" );
				}
				else if (c == 't'||c == 'T')
				{
					printf("Testing basic vector, matrix allocation and other support structures and functions...\n");
					test_vector_matrix();
					test_structure();
					test_functions();
				}
				else if (c == 27)
					break;
				else
					printf("Invalid key: %c = %d\n", c, c);
			}	// end if keyboard
		} // end for verbose loop

		cvDestroyAllWindows();

	} // end if verbose_flag
	else // if not verbose
	{
		// Do not wait for user input
		// Do not show images
		// Do the processing with the parameters in cpd.txt file
		
		ReadCPD( &cpar, "cpd_nv.txt" );
		//t0=-(clock_t) ((1.0/cmdlineparam.fps)* (double) CLOCKS_PER_SEC);
		for(;;)
		{		
			printf("FPS: %.2f\n", 1.0/ ((double) clock()/CLOCKS_PER_SEC-t0));
			t0 = (double) clock()/CLOCKS_PER_SEC;

			// Capture/load images 
			if (capture) // capture from video stream or file
			{
				frame = 0;
				frame = cvQueryFrame( capture );
				//cvCopy( frame, img, NULL);
				img = cvCloneImage( frame );

				if ((video_length>0) && (cvGetCaptureProperty( capture, CV_CAP_PROP_POS_FRAMES) == video_length))
					break;
			}
			else if (file_num > -1) // load image from file
			{
				makefilename(&cmdlineparam, filename, file_num);
				printf("Loading %s...\n", filename);
				img.load( filename, 0, CV_LOAD_IMAGE_COLOR);
				file_num++;
				if (file_num > cmdlineparam.fv)
					break;
			}

			// Process Input
			ProcessColorSegmentation(&img, cmdlineparam);

		} // end for not verbose loop

	} // end if not verbose_flag

	if (capture) cvReleaseCapture( &capture );

	if (cmdlineparam.verbose_flag)
	{
		printf("Done!\n");
		if (cmdlineparam.pause_flag)
		{
			printf("Press ESC to quit.\n");
			while( getch()!= 0x1B); // Wait till ESC is pressed
		}
	}

	return 0;
}