// ---------------------------------------------------------------------
// File: util.h
// Author: Miguel Torres-Torriti (c) 2008.11.21-2009.02.06 v. 2.0
// Change log:
//	2001.07.20, v.1.0, MTT: First version was created.
//  2009.02.06, v.2.0, MTT: Source code was improved.
// Description:
//  This file provides the following set of basic utility functions
//  (see details below and in .cpp file):
//    - Very basic utility functions
//      atob (convert ascci to boolean)
//      fatalError (print error message)
//    - Functions to check and swap byte order
//      (little endian or big endian architecture)
//    - Basic file openning and closing functions
//    - Functions to parse the command line arguments
//      modify the following functions according to your needs:
//        parsecmdline
//    - Functions to print help and welcome messages 
//      modify the following functions according to your needs:
//        printhelp, printwelcome
//    - Functions for loading and processing the program
//      environment file
//    - Some other functions for string parsing and manipulation
//    - Timing and delay functions
//
// Remarks:
//
//      WARNING: The high-resolution timers rely on windows.h
//      The set of functions timer_hr_* needs to be modified or
//      removed to compile under other OS different from Windows!
//
// ---------------------------------------------------------------------

#ifndef UTIL_H
#define UTIL_H

#include <stdio.h>                 /* Required by FILE */
#include "datastructures.h"

#define LSB_FIRST    0             /* Implies little-endian CPU... */
#define MSB_FIRST    1             /* Implies big-endian CPU...    */

#ifndef TRUE
#define TRUE    1
#endif
#ifndef FASLE
#define FALSE   0
#endif

#define BELL    '\x7'

 //#define BITSPERBYTE 8

 //#define CHARBITS    8
 //#define SHORTBITS   16
 //#define INTBITS     32
 //#define LONGBITS    32
 //#define PTRBITS     32
 //#define DOUBLEBITS  64
 //#define FLOATBITS   32

// --- BASIC UTILITY STRUCTURES --- //

// Structures required to handle the program options given as command line arguments.

typedef struct {			// Stores command line options set by the user
	char basefilename[256];		// Frame base filename
	char fileextension[250];	// Frame file extension
	char filenameformat[200];	// "%s%0<n>d.%s" (filename)(zero-padding number)(extension)
	long int iv;				// Frame initial value
	long int fv;				// Frame final value
	double fps;					// Video writer parameters
	char filename_out[512];		// Output filename
	bool output_flag;			// 1 = save results to disk
	bool display_flag;			// 1 = display results
	bool verbose_flag;			// 1 = print process status in the command line interface
	bool pause_flag;			// 1 = pause and wait for user to confirm actions
} cmdlinestruct;

extern cmdlinestruct cmdlineparam;

// --- BASIC UTILITY FUNCTIONS --- //

// --------------------------------------------------------------------------
// Very basic utility functions.
//
bool atob(char* str);			// Convert ASCII string "0" to bool 0,
								// otherwise bool 1
void fatalError (char *message);// Error message printing

// --------------------------------------------------------------------------
// Functions for checking the byte order (little endian or big endian)
// and swaping the byte order.  These can useful for reading raw data
// indpendently from the given CPU's byte order.
//
float          byteswap_SR_IR();
unsigned short byteswap_SUS_IUS();
int            CheckByteOrder();
void byteswap(unsigned char *ptr_in, int BytesPerWord);
void DisplayByteOrder(void);

// --------------------------------------------------------------------------
// Basic file openning and closing functions.
//
FILE *openFile (char *filename, char *options);
void closeFile (FILE *FP);

// --------------------------------------------------------------------------
// Functions for handling the program options given as command line arguments.
//
char *read_switch(int *argc, char **argv,
                  char *name, int args, char *defreturn); // Seeks and returns command line arguments as strings

int initcmdlinestruct(cmdlinestruct* cmdlineparam); // Initializes the cmdlinestruct structure
int printhelp(void); // Prints help message
int parsecmdline(int argc, char** argv, cmdlinestruct* cmdlineparam); // Parses the command line arguments
int printwelcome(void); // Prints the welcome message
int makefilename(cmdlinestruct* cmdlineparam, char* filename, int frame_number); // Builds the filename for sequences of file

// --------------------------------------------------------------------------
// Functions for loading and processing the program environment file.
//
char *loadenv(char *filename);
char *get_env_val(char *val, char *buffer, char *var_name, char *defreturn);
void regionParse(region *r, char *str);

// --------------------------------------------------------------------------
// Some other functions for string parsing and manipulation.
//
int parseString (char *string, char **tokenPtrs);
int parseStringDelim (char *string, char **tokenPtrs, char *delims);
char *stringUpper (char *string);

// --------------------------------------------------------------------------
// Timing functions.
//
void delay(double secs);
double timer_hr_count(void); // requires windows.h
double timer_hr_freq(void);  // requires windows.h
double timer_hr_elapsed(double initial_count); //requires windows.h

#endif